///###////////////////////////////////////////////////////////////////////////
//
// Burton Computer Corporation
// http://www.burton-computer.com
// http://www.cooldevtools.com
// $Id: FrequencyDB.cc,v 1.6 2002/09/17 21:39:36 bburton Exp $
//
// Copyright (C) 2000 Burton Computer Corporation
// ALL RIGHTS RESERVED
//
// This program is open source software; you can redistribute it
// and/or modify it under the terms of the Q Public License (QPL)
// version 1.0. Use of this software in whole or in part, including
// linking it (modified or unmodified) into other programs is
// subject to the terms of the QPL.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// Q Public License for more details.
//
// You should have received a copy of the Q Public License
// along with this program; see the file LICENSE.txt.  If not, visit
// the Burton Computer Corporation or CoolDevTools web site
// QPL pages at:
//
//    http://www.burton-computer.com/qpl.html
//    http://www.cooldevtools.com/qpl.html
//

#include <unistd.h>
#include <fcntl.h>
#include "FrequencyDB.h"
#include "FrequencyDBImpl_cache.h"

#ifdef USE_DBM

#include "FrequencyDBImpl_dbm.h"
typedef FrequencyDBImpl_dbm FrequencyDBImplClass_t;

#else
#ifdef USE_DB

#include "FrequencyDBImpl_bdb.h"
typedef FrequencyDBImpl_bdb FrequencyDBImplClass_t;

#else
#error Must define USE_DBM or USE_DB
#endif

#endif

static const string ID_PREFIX("__ID__");
static const string DIGEST_PREFIX("__MD5__");
static const string COUNT_WORD("__COUNT__");

FrequencyDB::FrequencyDB()
  : m_db(new FrequencyDBImpl_cache(new FrequencyDBImplClass_t))
{
  assert(m_db.get());
}

FrequencyDB::~FrequencyDB()
{
  close();
}

bool FrequencyDB::open(const string &filename)
{
  close();
  m_db->open(filename);
  return m_db.get() != 0;
}

void FrequencyDB::close()
{
  m_db->close();
}

void FrequencyDB::flush()
{
  m_db->flush();
}

void FrequencyDB::addWord(const string &word,
                          int good_count,
                          int spam_count)
{
  WordData counts;
  m_db->readWord(word, counts);

  counts.good_count = max(0, counts.good_count + good_count);
  counts.spam_count = max(0, counts.spam_count + spam_count);

  m_db->writeWord(word, counts);
}

void FrequencyDB::removeWord(const string &word,
                             int good_count,
                             int spam_count)
{
  addWord(word, -good_count, -spam_count);
}

void FrequencyDB::getWordCounts(const string &word,
                                int &good_count,
                                int &spam_count)
{
  WordData counts;
  m_db->readWord(word, counts);
  good_count = counts.good_count;
  spam_count = counts.spam_count;
}

void FrequencyDB::getMessageCounts(int &good_message_count,
                                   int &spam_message_count)
{
  getWordCounts(COUNT_WORD, good_message_count, spam_message_count);
}

bool FrequencyDB::containsMessage(const Message &msg,
                                  bool &is_spam)
{
  WordData counts;
  if (!m_db->readWord(DIGEST_PREFIX + msg.getDigest(), counts) &&
      !m_db->readWord(ID_PREFIX + msg.getID(), counts)) {
    is_spam = false;
    return false;
  }

  assert(counts.good_count == 0 || counts.good_count == 1);
  assert(counts.spam_count == 0 || counts.spam_count == 1);
  assert(counts.good_count + counts.spam_count <= 1);

  is_spam = (counts.spam_count > 0);
  return is_spam || (counts.good_count > 0);
}

void FrequencyDB::adjustWordCounts(const string &word,
                                   int delta,
                                   bool already_exists,
                                   bool is_spam)
{
  int other_delta = already_exists ? -delta : 0;
  if (is_spam) {
    addWord(word, other_delta, delta);
  } else {
    addWord(word, delta, other_delta);
  }
}

void FrequencyDB::addMessage(const Message &msg,
                             bool new_is_spam)
{
  assert(msg.getDigest().length() > 0);

  bool already_is_spam = false;
  bool already_exists = containsMessage(msg, already_is_spam);

  if (already_exists && (!already_is_spam == !new_is_spam)) {
    // already in database
    return;
  }

  for (int i = 0; i < msg.getTokenCount(); ++i) {
    const Token *tok = msg.getToken(i);
    adjustWordCounts(tok->getWord(), tok->getCount(), already_exists, new_is_spam);
  }
  if (msg.getID().length() > 0) {
    adjustWordCounts(ID_PREFIX + msg.getID(), 1, already_exists, new_is_spam);
  }
  adjustWordCounts(DIGEST_PREFIX + msg.getDigest(), 1, already_exists, new_is_spam);
  adjustWordCounts(COUNT_WORD, 1, already_exists, new_is_spam);

  if (is_debug) {
    if (already_exists) {
      cerr << "Updated message " << msg.getID()
           << "/" << msg.getDigest()
           << " in database." << endl;
    } else {
      cerr << "Added message " << msg.getID()
           << "/" << msg.getDigest()
           << " to database." << endl;
    }
  }
}

void FrequencyDB::removeMessage(const Message &msg)
{
  assert(msg.getDigest().length() > 0);

  bool is_spam = false;
  if (!containsMessage(msg, is_spam)) {
    // not in database
    return;
  }

  for (int i = 0; i < msg.getTokenCount(); ++i) {
    const Token *tok = msg.getToken(i);
    adjustWordCounts(tok->getWord(), -tok->getCount(), false, is_spam);
  }
  if (msg.getID().length() > 0) {
    adjustWordCounts(ID_PREFIX + msg.getID(), -1, false, is_spam);
  }
  adjustWordCounts(DIGEST_PREFIX + msg.getDigest(), -1, false, is_spam);
  adjustWordCounts(COUNT_WORD, -1, false, is_spam);

  if (is_debug) {
    cerr << "Removed message " << msg.getID()
         << "/" << msg.getDigest()
         << " from database." << endl;
  }
}

bool FrequencyDB::firstWord(string &word,
                            int &good_count,
                            int &spam_count) const
{
  WordData counts;
  bool found = m_db->firstWord(word, counts);
  if (found) {
    good_count = counts.good_count;
    spam_count = counts.spam_count;
  }
  return found;
}

bool FrequencyDB::nextWord(string &word,
                           int &good_count,
                           int &spam_count) const
{
  WordData counts;
  bool found = m_db->nextWord(word, counts);
  if (found) {
    good_count = counts.good_count;
    spam_count = counts.spam_count;
  }
  return found;
}

string FrequencyDB::getDatabaseType() const
{
  return m_db->getDatabaseType();
}
