///###////////////////////////////////////////////////////////////////////////
//
// Burton Computer Corporation
// http://www.burton-computer.com
// http://www.cooldevtools.com
// $Id: FrequencyDBImpl_bdb.cc,v 1.3 2002/09/19 16:15:37 bburton Exp $
//
// Copyright (C) 2000 Burton Computer Corporation
// ALL RIGHTS RESERVED
//
// This program is open source software; you can redistribute it
// and/or modify it under the terms of the Q Public License (QPL)
// version 1.0. Use of this software in whole or in part, including
// linking it (modified or unmodified) into other programs is
// subject to the terms of the QPL.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// Q Public License for more details.
//
// You should have received a copy of the Q Public License
// along with this program; see the file LICENSE.txt.  If not, visit
// the Burton Computer Corporation or CoolDevTools web site
// QPL pages at:
//
//    http://www.burton-computer.com/qpl.html
//    http://www.cooldevtools.com/qpl.html
//

#ifdef USE_DB

#include <unistd.h>
#include <fcntl.h>
#include "FrequencyDBImpl_bdb.h"

struct MyDBT : public DBT
{
  MyDBT()
  {
    memset(this, 0, sizeof(*this));
  }
};

FrequencyDBImpl_bdb::FrequencyDBImpl_bdb()
  : m_file(0),
    m_cursor(0)
{
}

FrequencyDBImpl_bdb::~FrequencyDBImpl_bdb()
{
  close();
}

bool FrequencyDBImpl_bdb::open(const string &filename)
{
  close();
  int ret = db_create(&m_file, NULL, 0);
  if (ret != 0) {
    m_file = 0;
    return false;
  }

  // SleepyCat in their infinite wisdom decided to change the open
  // function's signature in the 4.1 release.  Gee thanks for breaking
  // my code guys. That was real smart and so much better than
  // introducing a second open function rather than force me to embed
  // hideous ifdefs into my code.
#if DB_VERSION_MAJOR >= 4 && DB_VERSION_MINOR >= 1
  ret = m_file->open(m_file, NULL, filename.c_str(),
                     NULL, DB_BTREE, DB_CREATE, 0600);
#else
  ret = m_file->open(m_file, filename.c_str(),
                     NULL, DB_BTREE, DB_CREATE, 0600);
#endif
  if (ret != 0) {
    m_file = 0;
  }

  return ret != 0;
}

void FrequencyDBImpl_bdb::close()
{
  closeCursor();
  if (m_file) {
    m_file->close(m_file, 0);
    m_file = 0;
  }
}

void FrequencyDBImpl_bdb::flush()
{
}

void FrequencyDBImpl_bdb::writeWord(const string &word,
                                    const FrequencyDB::WordData &counts)
{
  assert(m_file);

  MyDBT key;
  key.data = (char *)word.c_str();
  key.size = word.length() + 1;

  MyDBT value;
  value.data = (char *)&counts;
  value.size = sizeof(counts);
  m_file->put(m_file, NULL, &key, &value, 0);
}

bool FrequencyDBImpl_bdb::readWord(const string &word,
                                   FrequencyDB::WordData &counts)
{
  assert(m_file);

  MyDBT key;
  key.data = (char *)word.c_str();
  key.size = word.length() + 1;
  return loadKey(key, counts);
}

void FrequencyDBImpl_bdb::closeCursor()
{
  if (m_cursor) {
    m_cursor->c_close(m_cursor);
    m_cursor = 0;
  }
}

bool FrequencyDBImpl_bdb::firstWord(string &word,
                                    FrequencyDB::WordData &counts)
{
  closeCursor();

  int ret = m_file->cursor(m_file, NULL, &m_cursor, 0);
  if (ret != 0) {
    return false;
  }
  assert(m_cursor);

  return nextWord(word, counts);
}

bool FrequencyDBImpl_bdb::nextWord(string &word,
                                   FrequencyDB::WordData &counts)
{
  if (!m_cursor) {
    return false;
  }

  MyDBT key;
  MyDBT value;
  int ret = m_cursor->c_get(m_cursor, &key, &value, DB_NEXT);
  if (ret != 0) {
    word.erase();
    counts.clear();
    closeCursor();
    return false;
  }

  if (!value.data || value.size != sizeof(FrequencyDB::WordData)) {
    word.erase();
    counts.clear();
    return false;
  }

  word.assign((const char *)key.data, max((u_int32_t)0, key.size - 1));
  counts = *((FrequencyDB::WordData *)value.data);
  return true;
}

bool FrequencyDBImpl_bdb::loadKey(DBT &key,
                                  string &word,
                                  FrequencyDB::WordData &counts) const
{
  if (key.data == NULL) {
    word.erase();
    counts.clear();
    return false;
  }

  if (key.size == 0) {
    word.erase();
  } else {
    word.assign((const char *)key.data, key.size - 1);
  }
  return loadKey(key, counts);
}

bool FrequencyDBImpl_bdb::loadKey(DBT &key,
                                  FrequencyDB::WordData &counts) const
{
  if (key.data == NULL) {
    counts.clear();
    return false;
  }

  MyDBT value;
  int ret = m_file->get(m_file, NULL, &key, &value, 0);
  if (ret != 0) {
    counts.clear();
    return false;
  }

  if (!value.data || value.size != sizeof(FrequencyDB::WordData)) {
    counts.clear();
    return false;
  }

  counts = *((FrequencyDB::WordData *)value.data);
  return true;
}

string FrequencyDBImpl_bdb::getDatabaseType() const
{
  return "BerkeleyDB-btree";
}

#endif // USE_DBM
