///###////////////////////////////////////////////////////////////////////////
//
// Burton Computer Corporation
// http://www.burton-computer.com
// http://www.cooldevtools.com
// $Id: MessageFactory.cc,v 1.19 2002/09/17 21:39:36 bburton Exp $
//
// Copyright (C) 2000 Burton Computer Corporation
// ALL RIGHTS RESERVED
//
// This program is open source software; you can redistribute it
// and/or modify it under the terms of the Q Public License (QPL)
// version 1.0. Use of this software in whole or in part, including
// linking it (modified or unmodified) into other programs is
// subject to the terms of the QPL.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// Q Public License for more details.
//
// You should have received a copy of the Q Public License
// along with this program; see the file LICENSE.txt.  If not, visit
// the Burton Computer Corporation or CoolDevTools web site
// QPL pages at:
//
//    http://www.burton-computer.com/qpl.html
//    http://www.cooldevtools.com/qpl.html
//

#include <strstream>
#include "Tokenizer.h"
#include "MessageFactory.h"

static const char *SKIPPED_HEADERS[] = {
  "X-SpamProbe",
  "X-Razor",
  "Status",
  "X-Status",
  "X-IMAP",
  0
};

MessageFactory::MessageFactory()
  : m_minWordLength(2),
    m_maxWordLength(90),
    m_replaceNonAsciiChars(true),
    m_nonAsciiChar('z'),
    m_removeHTML(true),
    m_useWordPairs(true),
    m_headersToInclude(NORMAL_HEADERS)
{
}

MessageFactory::~MessageFactory()
{
}

static bool is_skipped_header(const string &line)
{
  for (const char **s = SKIPPED_HEADERS; *s; ++s) {
    if (starts_with(line, *s)) {
      return true;
    }
  }
  return false;
}

void MessageFactory::addHeadersToMessage(Message &msg,
                                         MimeMessageReader &reader)
{
  if (m_headersToInclude == NORMAL_HEADERS) {
    addHeaderToMessage("reply-to", msg, reader);
    addHeaderToMessage("sender", msg, reader);
    addHeaderToMessage("originator", msg, reader);
    addHeaderToMessage("subject", msg, reader);
    addHeaderToMessage("subject", msg, reader); // add a second time for emphasis
    addHeaderToMessage("from", msg, reader);
    addHeaderToMessage("to", msg, reader);
    addHeaderToMessage("cc", msg, reader);
    addHeaderToMessage("received", msg, reader);

    string charset;
    if (reader.getCharSet(charset)) {
      addStringToMessage(charset, msg);
      addStringToMessage(charset, msg); // add a second time for emphasis
    }
  } else if (m_headersToInclude == NO_X_HEADERS) {
    string header;
    int header_count = reader.getFieldCount();
    for (int i = 0; i < header_count; ++i) {
      reader.getField(i, header);
      if (!starts_with(header, "X-") && !is_skipped_header(header)) {
        addStringToMessage(header, msg);
      }
    }
  } else if (m_headersToInclude != NO_HEADERS) {
    assert(m_headersToInclude == ALL_HEADERS);

    string header;
    int header_count = reader.getFieldCount();
    for (int i = 0; i < header_count; ++i) {
      reader.getField(i, header);
      if (!is_skipped_header(header)) {
        addStringToMessage(header, msg);
      }
    }
  }
}

void MessageFactory::initMessage(Message &msg,
                                 MimeMessageReader &reader)
{
  msg.clear();
  string value;
  if (reader.getField("message-id", value)) {
    msg.setID(value);
  }

  addHeadersToMessage(msg, reader);

  string text, content_type;
  while (reader.readText(text, content_type)) {
    addStringToMessage(content_type, msg);
    addTextToMessage(text, msg);
  }

  msg.setDigest(reader.getMD5Digest());

  if (is_debug) {
    cerr << "loaded message with digest " << msg.getDigest() << endl;
  }
}

void MessageFactory::addHeaderToMessage(const string &name,
                                        Message &msg,
                                        MimeMessageReader &reader)
{
  vector<string> text;
  if (reader.getField(name, text)) {
    for (vector<string>::const_iterator i = text.begin(); i != text.end(); ++i) {
      addStringToMessage(*i, msg);
    }
  }
}

void MessageFactory::addWordPartsToMessage(const string &word,
                                           Message &msg)
{
  const char *word_start = word.c_str();
  const char *s = word_start;
  while (*s) {
    while (*s && !is_alnum(*s) && !(*s & 0x80)) {
      ++s;
    }

    bool all_digits = true;
    const char *start = s;
    while (*s && (is_alnum(*s) || (*s & 0x80))) {
      all_digits = all_digits && is_digit(*s);
      ++s;
    }
    const char *end = s;

    if (!all_digits) {
      if (start != word_start) {
        msg.addToken(start);
      }

      if (((end - start) > 1) && *end && !all_digits) {
        msg.addToken(string(start, end));
      }
    }
  }
}

void MessageFactory::addWordToMessage(const string &word,
                                      Message &msg)
{
  assert(is_lower(word));

  if ((word.length() >= m_minWordLength)
      && (word.length() <= m_maxWordLength)
      && !is_digits(word)) {
    msg.addToken(word);
    if (m_useWordPairs && m_prevWord.length() > 0) {
      msg.addToken(m_prevWord + ' ' + word);
    }
    m_prevWord = word;
    addWordPartsToMessage(word, msg);
  } else {
    m_prevWord.erase();
  }
}

void MessageFactory::addStringToMessage(const string &value,
                                        Message &msg)
{
  m_prevWord.erase();

  Tokenizer tokenizer;
  for (const char *s = value.c_str(); *s; ++s) {
    char ch = (m_replaceNonAsciiChars && (*s & 0x80)) ? m_nonAsciiChar : *s;
    if (tokenizer.addChar(ch)) {
      addWordToMessage(tokenizer.getWord(), msg);
    }
  }

  if (tokenizer.stop()) {
    addWordToMessage(tokenizer.getWord(), msg);
  }
}

bool MessageFactory::isHTML(const string &text)
{
  return text.find("<html>") != string::npos ||
         text.find("<p>") != string::npos ||
         text.find("<br>") != string::npos ||
         text.find("<tr>") != string::npos ||
         text.find("<td>") != string::npos ||
         text.find("<font ") != string::npos ||
         text.find("<b>") != string::npos ||
         text.find("<a ") != string::npos;
}

void MessageFactory::addTextToMessage(const string &value,
                                      Message &msg)
{
  string lower = to_lower(value);
  if (isHTML(lower)) {
    if (m_removeHTML) {
      removeHTMLFromText(lower, msg);
    } else {
      expandEntitiesInHTML(lower);
    }
  }
  addStringToMessage(lower, msg);
}

static void add_entity(string &text,
                       const string &entity)
{
  if (entity == "amp") {
    text += '&';
  } else if (entity == "apos") {
    text += '\'';
  } else if (entity == "quot") {
    text += '"';
  } else if (entity == "lt") {
    text += '<';
  } else if (entity == "gt") {
    text += '>';
  } else if (entity == "nbsp") {
    text += ' ';
  } else if (entity[0] == '#') {
    int code = 0;
    istrstream in(entity.c_str() + 1);
    in >> code;
    text += (char)code;
  } else {
    text += entity;
  }
}

void MessageFactory::expandEntitiesInHTML(string &text)
{
  bool in_amp = false;
  string amp, new_text;
  for (const char *s = text.c_str(); *s; ++s) {
    if (in_amp) {
      if (*s == ';') {
        add_entity(new_text, amp);
        in_amp = false;
      } else {
        amp += *s;
      }
    } else if (*s == '&') {
      in_amp = true;
      amp.erase();
    } else {
      new_text += *s;
    }
  }
  text = new_text;
}

void MessageFactory::addHrefFromTag(const string &tag,
                                    Message &msg)
{
  const char *s = 0;
  string::size_type offset = tag.find(" href");
  if (offset != string::npos) {
    s = tag.c_str() + offset + 5;
  } else {
    offset = tag.find(" src");
    if (offset != string::npos) {
      s = tag.c_str() + offset + 4;
    }
  }

  if (!s) {
    // no url attribute found
    return;
  }

  while (*s && is_space(*s)) {
    ++s;
  }
  if (*s != '=') {
    // no value for attribute
    return;
  }

  while (*s && is_space(*s)) {
    ++s;
  }

  string url;
  if (*s == '"' || *s == '\'') {
    char quote = *s;
    while (*s && *s != quote) {
      url += *s;
      ++s;
    }
  } else {
    while (*s && !is_space(*s)) {
      url += *s;
      ++s;
    }
  }

  expandEntitiesInHTML(url);

  if (is_debug) {
    cerr << "HREF " << url << endl;
  }

  addStringToMessage(url, msg);
}

void MessageFactory::removeHTMLFromText(string &text,
                                        Message &msg)
{
  string new_text;
  string amp, tag;
  bool in_tag = false;
  bool in_amp = false;
  for (const char *s = text.c_str(); *s; ++s) {
    if (in_tag) {
      if (*s == '>') {
        in_tag = false;
        addHrefFromTag(tag, msg);
      } else {
        tag += *s;
      }
    } else if (*s == '<') {
      in_tag = true;
      in_amp = false;
      tag.erase();
    } else if (*s == '&') {
      in_amp = true;
      amp.erase();
    } else if (in_amp) {
      if (*s == ';') {
        add_entity(new_text, amp);
        in_amp = false;
      } else {
        amp += *s;
      }
    } else {
      new_text += *s;
    }
  }
  text = new_text;
}
