///###////////////////////////////////////////////////////////////////////////
//
// Burton Computer Corporation
// http://www.burton-computer.com
// http://www.cooldevtools.com
// $Id: SpamFilter.cc,v 1.12 2002/09/19 16:15:38 bburton Exp $
//
// Copyright (C) 2000 Burton Computer Corporation
// ALL RIGHTS RESERVED
//
// This program is open source software; you can redistribute it
// and/or modify it under the terms of the Q Public License (QPL)
// version 1.0. Use of this software in whole or in part, including
// linking it (modified or unmodified) into other programs is
// subject to the terms of the QPL.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// Q Public License for more details.
//
// You should have received a copy of the Q Public License
// along with this program; see the file LICENSE.txt.  If not, visit
// the Burton Computer Corporation or CoolDevTools web site
// QPL pages at:
//
//    http://www.burton-computer.com/qpl.html
//    http://www.cooldevtools.com/qpl.html
//

#include <memory>
#include <stdexcept>
#include <algorithm>
#include "SpamFilter.h"

static const double NEARLY_ZERO = 0.0000001;
static const int MAX_COUNT_FOR_SORT = 5;

typedef vector<Token *>::size_type vtokindex_t;

SpamFilter::SpamFilter()
  : m_goodBias(2),
    m_termsForScore(27),
    m_maxWordRepeats(2),
    m_minWordCount(5),
    m_newWordScore(0.3)
{
}

SpamFilter::~SpamFilter()
{
  close();
}

void SpamFilter::open(const File &db_file)
{
  if (!m_db.open(db_file.getPath())) {
    throw runtime_error("unable to open words database");
  }
}

double SpamFilter::computeRatio(int count,
                                int total_count)
{
  if (count >= total_count) {
    return 1.0;
  }

  return (double)count / (double)total_count;
}

double SpamFilter::scoreToken(Token *tok,
                              int min_count,
                              int good_message_count,
                              int spam_message_count)
{
  assert(tok->getCount() > 0);

  int good_count, spam_count;
  m_db.getWordCounts(tok->getWord(), good_count, spam_count);
  good_count *= m_goodBias;

  int count = good_count + spam_count;

  if (count < 0 && is_debug) {
    cerr << "WARNING: token " << tok->getWord() << " has negative count" << endl;
  }

  if (count < min_count) {
    return m_newWordScore;
  }

  double score = 0.5;
  double good_ratio = computeRatio(good_count, good_message_count);
  double spam_ratio = computeRatio(spam_count, spam_message_count);
  double denom = good_ratio + spam_ratio;
  if (denom > 0.0) {
    score = spam_ratio / denom;
  }

  return max(0.00001, min(0.99999, score));
}

inline bool non_zero(double v)
{
  if (v > 0) {
    return v > NEARLY_ZERO;
  } else {
    return v < -NEARLY_ZERO;
  }
}

static int token_qsort_criterion(Token **t1p, Token **t2p)
{
  Token *t1 = *t1p;
  Token *t2 = *t2p;

  double dist = t1->getDistanceFromMean() - t2->getDistanceFromMean();
  if (non_zero(dist)) {
    return (dist > 0) ? -1 : 1;
  }

  // Break ties with word count if possible.  Use a limit on count to
  // impose some sanity when the counts are ridiculously big.
  int diff = min(MAX_COUNT_FOR_SORT, t1->getCount()) - min(MAX_COUNT_FOR_SORT, t2->getCount());
  if (diff != 0) {
    return -diff;
  }

  // next use lower (less spammy) score to break tie if possible
  dist = t1->getScore() - t2->getScore();
  if (non_zero(dist)) {
    return (dist > 0) ? 1 : -1;
  }

  // as a last resort sort alphabetically
  return t1->getWord().compare(t2->getWord());
}

void SpamFilter::getSortedTokens(const Message &msg,
                                 int max_tokens,
                                 vector<Token *> &tokens)
{
  int good_message_count, spam_message_count;
  m_db.getMessageCounts(good_message_count, spam_message_count);

  tokens.clear();
  for (int i = 0; i < msg.getTokenCount(); ++i) {
    Token *tok = msg.getToken(i);
    assert(tok);

    double score = scoreToken(tok, m_minWordCount, good_message_count, spam_message_count);
    if (score > 0.0) {
      tok->setScore(score);
      tokens.push_back(tok);
    } else {
      tok->setScore(0.0);
    }
  }

  if (tokens.size() == 0) {
    return;
  }

  // STL sort is bugged so we have to use qsort!
  vtokindex_t num_tokens = tokens.size();
  NewPtr<Token *> sorted(new Token *[num_tokens]);
  for (vtokindex_t i = 0; i < num_tokens; ++i) {
    sorted.get()[i] = tokens[i];
  }

  qsort(sorted.get(), num_tokens, sizeof(Token *),
        (int (*) (const void *, const void *))token_qsort_criterion);

  if (is_debug) {
    for (vtokindex_t i = 0; i < num_tokens; ++i) {
      Token *tok = sorted.get()[i];
      cerr << "SORTED " << tok->getWord()
           << " count " << tok->getCount()
           << " score " << tok->getScore()
           << " diff " << tok->getDistanceFromMean()
           << endl;
    }
  }

  tokens.clear();
  vtokindex_t max_size = (vtokindex_t)max_tokens;
  for (vtokindex_t i = 0; i < max_size && i < num_tokens; ++i) {
    tokens.push_back(sorted.get()[i]);
  }
}

double SpamFilter::scoreMessage(Message &msg)
{
  vector<Token *> tokens;
  getSortedTokens(msg, m_termsForScore, tokens);

  msg.clearTopTokens();

  double outer = 1.0;
  double inner = 1.0;
  int remaining = m_termsForScore;
  for (vector<Token *>::size_type i = 0; i < tokens.size() && remaining > 0; ++i) {
    int count = tokens[i]->getCount();
    double score = tokens[i]->getScore();

    msg.addTopToken(tokens[i]);

    assert(count > 0);
    assert(remaining > 0);

    int times = min(remaining, min(m_maxWordRepeats, count));
    if (is_debug) {
      cerr << "** TOKEN " << tokens[i]->getWord() << ": score " << score << ": times " << times << endl;
    }

    remaining -= times;
    while (times-- > 0) {
      inner = inner * score;
      outer = outer * (1.0 - score);
    }
  }
  
  if (is_debug) {
    cerr << "** INNER " << inner << ": OUTER " << outer << ": SCORE "
         << (inner / (inner + outer)) << endl;
  }

  return inner / (inner + outer);
}
