///###////////////////////////////////////////////////////////////////////////
//
// Burton Computer Corporation
// http://www.burton-computer.com
// http://www.cooldevtools.com
// $Id: spamprobe.cc,v 1.26 2002/09/19 18:53:51 bburton Exp $
//
// Copyright (C) 2000 Burton Computer Corporation
// ALL RIGHTS RESERVED
//
// This program is open source software; you can redistribute it
// and/or modify it under the terms of the Q Public License (QPL)
// version 1.0. Use of this software in whole or in part, including
// linking it (modified or unmodified) into other programs is
// subject to the terms of the QPL.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// Q Public License for more details.
//
// You should have received a copy of the Q Public License
// along with this program; see the file LICENSE.txt.  If not, visit
// the Burton Computer Corporation or CoolDevTools web site
// QPL pages at:
//
//    http://www.burton-computer.com/qpl.html
//    http://www.cooldevtools.com/qpl.html
//

#include <unistd.h>
#include <fstream>
#include <stdexcept>
#include <cstdio>
#include <iomanip>
#include "File.h"
#include "LockFile.h"
#include "SpamFilter.h"
#include "MessageFactory.h"
#include "MimeMessageReader.h"

static const string VERSION("0.7c");

static const string DB_FILENAME("sp_words");

static int MESSAGES_PER_FLUSH = 250;
static const int MAX_MESSAGES_PER_FLUSH = 100000;

static const string CMD_RECEIVE("receive");
static const string CMD_GOOD("good");
static const string CMD_DUMP("dump");
static const string CMD_IMPORT("import");
static const string CMD_EXPORT("export");
static const string CMD_SPAM("spam");
static const string CMD_SCORE("score");
static const string CMD_REMOVE("remove");
static const string CMD_FIND_SPAM("find-spam");
static const string CMD_FIND_GOOD("find-good");

static const double SPAM_THRESHOLD = 0.9;

static const char *print_score(double score)
{
  static char buffer[30];
  sprintf(buffer, "%9.7f", score);
  return buffer;
}

void print_terms(ostream &out,
                 Message &msg,
                 const string &indent)
{
  out << indent << "Spam Prob   Count  Word" << endl;
  for (int i = 0; i < msg.getTopTokenCount(); ++i) {
    Token *tok = msg.getTopToken(i);
    out << indent
        << print_score(tok->getScore())
        << setw(8) << tok->getCount()
        << "  " << tok->getWord()
        << endl;
  }
}

void find_message(SpamFilter &filter,
                  Message &msg,
                  int message_num,
                  bool find_spam,
                  bool show_terms)
{
  double score = filter.scoreMessage(msg);
  bool is_spam = (score >= SPAM_THRESHOLD);
  if (is_spam == find_spam) {
    cout << "Message Num: " << message_num << endl
         << "Message-ID: " << msg.getID() << endl
         << "Digest: " << msg.getDigest() << endl
         << "Score: "
         << print_score(score)
         << endl;
    if (show_terms && msg.getTopTokenCount() > 0) {
      print_terms(cout, msg, "       ");
    }
  }
}

void dump_words(const File &basedir,
                SpamFilter &filter)
{
  LockFile lock(File(basedir, "lock"));
  lock.lock(LockFile::EXCLUSIVE);
  filter.open(File(basedir, DB_FILENAME));

  string word;
  int good_count, spam_count;
  bool again = filter.getDB()->firstWord(word, good_count, spam_count);
  while (again) {
    cout << setw(10) << good_count
         << setw(10) << spam_count
         << "  "
         << word
         << endl;
    again = filter.getDB()->nextWord(word, good_count, spam_count);
  }

  filter.close();
}

void export_words(const File &basedir,
                  SpamFilter &filter)
{
  LockFile lock(File(basedir, "lock"));
  lock.lock(LockFile::EXCLUSIVE);
  filter.open(File(basedir, DB_FILENAME));

  string word;
  int good_count, spam_count;
  bool again = filter.getDB()->firstWord(word, good_count, spam_count);
  while (again) {
    cout << good_count
         << ','
         << spam_count
         << ",\""
         << word
         << "\""
         << endl;
    again = filter.getDB()->nextWord(word, good_count, spam_count);
  }

  filter.close();
}

void import_words(const File &basedir,
                  SpamFilter &filter,
                  istream &in)
{
  LockFile lock(File(basedir, "lock"));
  lock.lock(LockFile::EXCLUSIVE);
  filter.open(File(basedir, DB_FILENAME));

  int record_count = 0;
  MimeLineReader reader(in);
  while (reader.readLine()) {
    if (is_debug) {
      cerr << "IMPORT " << reader.getLine() << endl;
    }
    string::size_type comma1 = reader.getLine().find(',');
    if (comma1 == string::npos) {
      continue;
    }
    string::size_type comma2 = reader.getLine().find(',', comma1 + 1);
    if (comma2 == string::npos) {
      continue;
    }
    int good_count = atoi(reader.getLine().substr(0, comma1).c_str());
    int spam_count = atoi(reader.getLine().substr(comma1 + 1, comma2 - comma1 - 1).c_str());
    string word(reader.getLine(), comma2 + 1, reader.getLine().size());
    if (word.length() < 1) {
      continue;
    }
    if (word[0] == '"') {
      filter.getDB()->addWord(word.substr(1, word.length() - 2), good_count, spam_count);
    } else {
      filter.getDB()->addWord(word, good_count, spam_count);
    }
    if (++record_count > 10000) {
      filter.flush();
      record_count = 0;
    }
  }
  filter.close();
}

void process_stream(const File &basedir,
                    SpamFilter &filter,
                    istream &in_stream,
                    const string &command,
                    bool ignore_from,
                    bool ignore_content_length,
                    bool show_terms,
                    MessageFactory &factory)
{
  LockFile lock(File(basedir, "lock"));
  lock.lock(LockFile::EXCLUSIVE);
  filter.open(File(basedir, DB_FILENAME));

  Message msg;

  int message_num = 0;
  MimeMessageReader in(in_stream, ignore_from, ignore_content_length);
  while (in.readNextHeader()) {
    ++message_num;
    if (is_debug) {
      cerr << "*** NEW MESSAGE ***" << endl;
    }
    factory.initMessage(msg, in);
    if (command == CMD_GOOD) {
      filter.ensureGoodMessage(msg);
    } else if (command == CMD_SPAM) {
      filter.ensureSpamMessage(msg);
    } else if (command == CMD_REMOVE) {
      filter.removeMessage(msg);
    } else if (command == CMD_RECEIVE) {
      double score = filter.scoreMessage(msg);
      if (score >= SPAM_THRESHOLD) {
        cout << "SPAM ";
        filter.ensureSpamMessage(msg);
      } else {
        cout << "GOOD ";
        filter.ensureGoodMessage(msg);
      }
      cout << print_score(score)
           << endl;
    } else if (command == CMD_FIND_SPAM) {
      find_message(filter, msg, message_num, true, show_terms);
    } else if (command == CMD_FIND_GOOD) {
      find_message(filter, msg, message_num, false, show_terms);
    } else if (command == CMD_SCORE) {
      cout << print_score(filter.scoreMessage(msg))
           << endl;
      if (show_terms && msg.getTopTokenCount() > 0) {
        print_terms(cout, msg, "    ");
      }
    } else {
      cerr << "error: unknown command " << command << endl;
      exit(1);
    }
    if (is_debug) {
      cerr << "*** END OF MESSAGE ***" << endl;
    }

    if (message_num % MESSAGES_PER_FLUSH == 0) {
      filter.flush();
    }
  }

  filter.close();
}

bool process_test_case(const string &test_case,
                       MessageFactory &factory,
                       SpamFilter &filter)
{
  bool failed = false;

  if (test_case.length() > 0) {
    if (test_case == "graham") {
      factory.setUseWordPairs(false);
      factory.setRemoveHTML(false);
      factory.setMinWordLength(1);
      factory.setMaxWordLength(90);
      filter.setTermsForScore(15);
      filter.setMaxWordRepeats(1);
      filter.setNewWordScore(0.4);
      factory.setHeadersToInclude(MessageFactory::ALL_HEADERS);
    } else if (test_case == "old-graham") {
      factory.setUseWordPairs(false);
      factory.setRemoveHTML(false);
      factory.setMinWordLength(1);
      factory.setMaxWordLength(90);
      filter.setTermsForScore(15);
      filter.setMaxWordRepeats(1);
      filter.setNewWordScore(0.2);
      factory.setHeadersToInclude(MessageFactory::ALL_HEADERS);
    } else if (test_case == "wide-open") {
      factory.setRemoveHTML(false);
      factory.setHeadersToInclude(MessageFactory::ALL_HEADERS);
    } else if (test_case == "low-score") {
      filter.setNewWordScore(0.2);
    } else if (test_case == "no-phrase") {
      factory.setUseWordPairs(false);
    } else if (test_case != "normal") {
      cerr << "error: unknown test case: " << test_case << endl;
      failed = true;
    }
  }

  return failed;
}

static bool parse_int_arg(const char *arg,
                          const char *option,
                          int &value,
                          int min_val,
                          int max_val)
{
  value = atoi(arg);
  if (value >= min_val && value <= max_val) {
    return true;
  }

  cerr << "error: -" << option << " requires integer argument between "
       << min_val << " and " << max_val << endl;
  return false;
}

static bool set_headers(MessageFactory &factory,
                        const char *optarg)
{
  bool successful = true;
  if (strcmp(optarg, "all") == 0) {
    factory.setHeadersToInclude(MessageFactory::ALL_HEADERS);
  } else if (strcmp(optarg, "nox") == 0) {
    factory.setHeadersToInclude(MessageFactory::NO_X_HEADERS);
  } else if (strcmp(optarg, "normal") == 0) {
    factory.setHeadersToInclude(MessageFactory::NORMAL_HEADERS);
  } else if (strcmp(optarg, "none") == 0) {
    factory.setHeadersToInclude(MessageFactory::NO_HEADERS);
  } else {
    cerr << "error: -H option requires all, nox, none, or normal" << endl;
    successful = false;
  }
  return successful;
}

static void print_version(SpamFilter &filter)
{
  cout << "SpamProbe v" << VERSION
       << " using " << filter.getDB()->getDatabaseType()
       << " database." << endl
       << endl
       << "Copyright 2002 Burton Computer Corporation" << endl
       << "This program is distributed in the hope that it will be useful," << endl
       << "but WITHOUT ANY WARRANTY; without even the implied warranty of" << endl
       << "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the" << endl
       << "Q Public License for more details." << endl;
}

int main(int argc,
         char **argv)
{
  try {
    int opt;
    string test_case = "";
    bool usage_error = false;
    bool force_mbox = false;
    bool create_dir = false;
    bool ignore_from = false;
    bool ignore_content_length = false;
    bool show_terms = false;
    File basedir(File::getHomeDir(), ".spamprobe");

    MessageFactory factory;
    factory.setReplaceNonAsciiChars('Z');

    SpamFilter filter;

    optind = 1;
    int opt_value;
    while ((opt = getopt(argc, argv, "a:cd:hH:mr:s:t:TvVw:Y78")) != EOF) {
      switch (opt) {
      case 'a':
        if (strlen(optarg) != 1) {
          cerr << "error: -a option requires a one character argument" << endl;
          usage_error = true;
        } else {
          factory.setReplaceNonAsciiChars(optarg[0]);
        }
        break;

      case 'c':
        create_dir = true;
        break;

      case 'd':
        basedir.setPath(optarg);
        break;

      case 'h':
        factory.setRemoveHTML(false);
        break;

      case 'H':
        if (!set_headers(factory, optarg)) {
          usage_error = true;
        }
        break;

      case 'm':
        force_mbox = true;
        break;

      case 'r':
        if (parse_int_arg(optarg, "r", opt_value, 1, 10)) {
          filter.setMaxWordRepeats(opt_value);
        } else {
          usage_error = true;
        }
        break;

      case 's':
        if (parse_int_arg(optarg, "s", opt_value, 0, MAX_MESSAGES_PER_FLUSH)) {
          if (opt_value == 0) {
            opt_value = MAX_MESSAGES_PER_FLUSH;
          }
          MESSAGES_PER_FLUSH = opt_value;
        } else {
          usage_error = true;
        }
        break;

      case 't':
        test_case = optarg;
        break;

      case 'T':
        show_terms = true;
        break;

      case 'v':
        is_debug = true;
        break;

      case 'V':
        print_version(filter);
        return 0;

      case 'w':
        if (parse_int_arg(optarg, "w", opt_value, 5, 500)) {
          filter.setTermsForScore(opt_value);
        } else {
          usage_error = true;
        }
        break;

      case 'Y':
        ignore_content_length = true;
        break;

      case '7':
        factory.setReplaceNonAsciiChars(' ');
        break;

      case '8':
        factory.setReplaceNonAsciiChars(-1);
        break;

      default:
        usage_error = true;
        break;
      }
    }

    if (process_test_case(test_case, factory, filter)) {
      usage_error = true;
    }

    if (!basedir.isDirectory()) {
      if (create_dir) {
        basedir.makeDirectory(0700);
      } else {
        cerr << "error: " << basedir.getPath() << " is not a directory" << endl;
        usage_error = true;
      }
    }
    if (usage_error || (argc - optind) < 1) {
      cerr << "usage: spamprobe [-chmTvVY78] a char -d basedir -H headers command [filename...]"
           << endl;
      return 1;
    }

    string command(argv[optind++]);

    if (command == CMD_RECEIVE && !force_mbox) {
      ignore_from = true;
      ignore_content_length = true;
    }

    if (command == CMD_DUMP) {
      dump_words(basedir, filter);
      return 0;
    }

    if (command == CMD_EXPORT) {
      export_words(basedir, filter);
      return 0;
    }

    if (optind == argc) {
      if (command == CMD_IMPORT) {
        import_words(basedir, filter, cin);
      } else {
        process_stream(basedir, filter, cin, command,
                       ignore_from, ignore_content_length, show_terms,
                       factory);
      }
      return 0;
    }

    for (int i = optind; i < argc; ++i) {
      File mbox(argv[i]);
      if (!mbox.isFile()) {
        cerr << "error: skipped " << mbox.getPath() << ": file not found" << endl;
      } else {
        LockFile lock(mbox);
        lock.lock(LockFile::SHARED);

        ifstream in(argv[i]);
        if (command == CMD_IMPORT) {
          import_words(basedir, filter, in);
        } else {
          process_stream(basedir, filter, in, command,
                         ignore_from, ignore_content_length, show_terms,
                         factory);
        }
      }
    }
  } catch (runtime_error &ex) {
    cerr << "caught runtime exception: " << ex.what() << endl;
    return 1;
  } catch (logic_error &ex) {
    cerr << "caught logic exception: " << ex.what() << endl;
    return 1;
  } catch (...) {
    cerr << "caught unknown exception" << endl;
    return 1;
  }

  return 0;
}
